/*
 * ps3mca-tool - PlayStation 3 Memory Card Adaptor Software
 * Copyright (C) 2011 - jimmikaelkael <jimmikaelkael@wanadoo.fr>
 * Copyright (C) 2011 - "someone who wants to stay anonymous"
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <inttypes.h>

#include "usbd.h"
#include "mcio.h"

#define PROGRAM_NAME	"PS3MCA-TOOL"
#define PROGRAM_VER	"1.3"

#define CMD_NONE	0
#define CMD_MCINFO	1
#define CMD_MCFREE	2
#define CMD_MCIMG	3
#define CMD_MCUNFORMAT	4
#define CMD_MCFORMAT	5
#define CMD_LIST	6
#define CMD_EXTRACT	7
#define CMD_INJECT	8
#define CMD_MKDIR	9
#define CMD_RMDIR	10
#define CMD_REMOVE	11
#define CMD_CROSSLINK	12
#define CMD_MCCK	13
#define CMD_MCKELF	14

char* ps2boot_region = "AEI";
char* ps2boot_files_tab[]= {
	"/BREXEC-SYSTEM/osdsys.elf",
	"/BREXEC-SYSTEM/osd110.elf",
	"/BREXEC-SYSTEM/osd120.elf",
	"/BREXEC-SYSTEM/osd130.elf",
	"/BREXEC-SYSTEM/osdmain.elf",
	NULL
};

static void print_usage(int argc, char **argv)
{
	printf("Copyright (C) 2011 - jimmikaelkael & 'someone who wants to stay anonymous'\n");
	printf("\n");
	printf("Usage:\n");
	printf("%s <command> [<arguments>]\n", argv[0]);
	printf("\n");
	printf("Available commands:\n");
	printf("\t --mc-info, -i\n");
	printf("\t --mc-free, -f\n");
	printf("\t --mc-image, -img <output filepath>\n");
	printf("\t --mc-unformat\n");
	printf("\t --mc-format\n");
	printf("\t --list, -ls <mc path>\n");
	printf("\t --extract-file, -x <mc filepath> <output filepath>\n");
	printf("\t --inject-file, -in <input filepath> <mc filepath>\n");
	printf("\t --make-directory, -mkdir <mc path>\n");
	printf("\t --remove-directory, -rmdir <mc path>\n");
	printf("\t --remove, -rm <mc filepath>\n");
	printf("\t --file-crosslink, -cl <real mc filepath> <dummy mc filepath>\n");
	printf("\t --content-key, -ck <disk kelf filepath>\n");
	printf("\t --sign-kelf, -k <disk kelf filepath> <card kelf filepath>\n");
	printf("\n");
}

static int cmd_mcinfo(void)
{
	int r;
	int pagesize, blocksize, cardsize, cardflags;

	r = mcio_mcGetInfo(&pagesize, &blocksize, &cardsize, &cardflags);
	if (r < 0)
		return r;

	printf("PS2 Memory Card Informations\n");
	printf("Page size:  %d bytes\n", pagesize);
	printf("Block size: %d pages\n", blocksize);
	printf("MC size:    %d MB\n", cardsize / 1024 / 1024);
	if (cardflags & 1)
		printf("MC claims to support ECC\n");
	if (cardflags & 8)
		printf("MC claims to support bad blocks management\n");
	printf("erase byte: 0x%02X\n", (cardflags & 16) ? 0x00 : 0xff);

	return 0;
}

static int cmd_mcfree(void)
{
	int r;
	int cardfree;

	printf("PS2 Memory Card free space\n");
	printf("Calculating free space...\n");

	r = mcio_mcGetAvailableSpace(&cardfree);
	if (r < 0)
		return r;

	printf("Available space: %d KB\n", cardfree / 1024);

	return 0;
}

static int cmd_mcimg(char *output)
{
	int r, i;
	int pagesize, blocksize, cardsize, cardflags;

	r = mcio_mcGetInfo(&pagesize, &blocksize, &cardsize, &cardflags);
	if (r < 0)
		return -1;

	FILE *fh = fopen(output, "wb");
	if (fh == NULL)
		return -2;

	void *buf = malloc(pagesize);
	if (buf == NULL) {
		fclose(fh);
		return -3;
	}

	for (i = 0; i < (cardsize / pagesize); i++) {
		mcio_mcReadPage(i, buf);
		r = fwrite(buf, 1, pagesize, fh);
		if (r != pagesize) {
			free(buf);
			fclose(fh);
			return -4;
		}
	}

	fclose(fh);
	free(buf);

	return 0;
}

static int cmd_mcunformat(void)
{
	int r;

	printf("PS2 Memory Card unformat\n");
	printf("Unformating MC...\n");

	r = mcio_mcUnformat();
	if (r < 0)
		return r;

	printf("Memory card succesfully unformated.\n");

	return 0;
}

static int cmd_mcformat(void)
{
	int r;

	printf("PS2 Memory Card format\n");
	printf("Formating MC...\n");

	r = mcio_mcFormat();
	if (r < 0)
		return r;

	printf("Memory card succesfully formated.\n");

	return 0;
}

static int cmd_list(char *path)
{
	int r, fd;

	fd = mcio_mcDopen(path);
	if (fd >= 0) {
		struct io_dirent dirent;
		printf("---------- Filename ----------  |  Type  | Last Modification (UTC)\n");
		do {
			r = mcio_mcDread(fd, &dirent);
			if ((r)) { /* && (strcmp(dirent.name, ".")) && (strcmp(dirent.name, ".."))) { */
				int tabnum = (32/7) - (strlen(dirent.name) / 7);
				printf("%s ", dirent.name);
				int i;
				for (i=0; i<tabnum; i++)
					printf("\t");
				if (!(dirent.stat.mode & sceMcFileAttrSubdir))
					printf("| <file> | ");
				else
					printf("| <dir>  | ");
				printf("%02d/%02d/%04d-", dirent.stat.mtime.Month, dirent.stat.mtime.Day, dirent.stat.mtime.Year);
				printf("%02d:%02d:%02d", dirent.stat.mtime.Hour, dirent.stat.mtime.Min, dirent.stat.mtime.Sec);
				printf("\n");
			}
		} while (r);

		mcio_mcDclose(fd);
	}

	return fd;
}

static int cmd_extract(char *filepath, char *output)
{
	int fd, r;

	printf("Reading file: '%s'...\n", filepath);

	fd = mcio_mcOpen(filepath, sceMcFileAttrReadable | sceMcFileAttrFile);
	if (fd < 0)
		return fd;

	int filesize = mcio_mcSeek(fd, 0, SEEK_END);
	mcio_mcSeek(fd, 0, SEEK_SET);
	uint8_t *p = malloc(filesize);
	if (p == NULL)
		return -1000;

	r = mcio_mcRead(fd, p, filesize);
	if (r != filesize) {
		mcio_mcClose(fd);
		free(p);
		return -1001;
	}

	mcio_mcClose(fd);

	FILE *fh = fopen(output, "wb");
	if (fh == NULL) {
		free(p);
		return -1002;
	}

	printf("Writing data to: '%s'...\n", output);

	r = fwrite(p, 1, filesize, fh);
	if (r != filesize) {
		fclose(fh);
		free(p);
		return -1003;
	}

	fclose(fh);

	free(p);

	return fd;
}

static int cmd_inject(char *input, char *filepath)
{
	int fd, r;

	FILE *fh = fopen(input, "rb");
	if (fh == NULL)
		return -1000;

	fseek(fh, 0, SEEK_END);
	int filesize = ftell(fh);
	if (!filesize) {
		fclose(fh);
		return -1001;
	}
	fseek(fh, 0, SEEK_SET);

	printf("Reading file: '%s'...\n", input);

	uint8_t *p = malloc(filesize);
	if (p == NULL) {
		return -1002;
	}

	r = fread(p, 1, filesize, fh);
	if (r != filesize) {
		fclose(fh);
		free(p);
		return -1003;
	}

	fclose(fh);

	printf("Writing data to: '%s'...\n", filepath);

	fd = mcio_mcOpen(filepath, sceMcFileCreateFile | sceMcFileAttrWriteable | sceMcFileAttrFile);
	if (fd < 0) {
		free(p);
		return fd;
	}

	r = mcio_mcWrite(fd, p, filesize);
	if (r != filesize) {
		mcio_mcClose(fd);
		free(p);
		return -1004;
	}

	mcio_mcClose(fd);

	free(p);

	return fd;
}

static int cmd_mkdir(char *path)
{
	printf("Creating directory: '%s'...\n", path);
	return mcio_mcMkDir(path);
}

static int cmd_rmdir(char *path)
{
	printf("Removing directory: '%s'...\n", path);
	return mcio_mcRmDir(path);
}

static int cmd_remove(char *path)
{
	printf("Removing file: '%s'...\n", path);
	return mcio_mcRemove(path);
}

static int cmd_crosslink(char *real_filepath, char *dummy_filepath)
{
	printf("Cross-linking file: '%s' to '%s'...\n", dummy_filepath, real_filepath);
	mcio_mcCreateCrossLinkedFile(real_filepath, dummy_filepath);
	return 0;
}

static int cmd_mcck(char *disk_kelf_path)
{
	int i;
	uint8_t ContentKey[32];

	FILE *fh = fopen(disk_kelf_path, "rb");
	if (fh == NULL)
		return -1;

	uint8_t *buf = malloc(0xffff);
	if (buf == NULL)
		return -2;

	int r = fread(buf, 1, 0xffff, fh);
	if (r < 0)
		return -3;

	fclose(fh);

	mcio_mcEncryptContentKey(buf, ContentKey);

	printf("PS2 Memory Card Content Key\n");
	printf("target Kelf file: '%s'\n", disk_kelf_path);
	printf("Kbit: ");
	for (i=0; i<16; i++)
		printf("%02x ", ContentKey[i]);
	printf("\n");
	printf("Kc  : ");
	for (i=0; i<16; i++)
		printf("%02x ", ContentKey[i+16]);
	printf("\n");

	free(buf);

	return 0;
}

static int cmd_mckelf(char *disk_kelf_path, char *card_kelf_path)
{
	int i;
	uint8_t ContentKey[32];

	FILE *fh = fopen(disk_kelf_path, "rb");
	if (fh == NULL)
		return -1;

	fseek(fh, 0, SEEK_END);
	int filesize = ftell(fh);
	if (!filesize) {
		fclose(fh);
		return -2;
	}
	fseek(fh, 0, SEEK_SET);

	uint8_t *buf = malloc(filesize);
	if (buf == NULL)
		return -3;

	int r = fread(buf, 1, filesize, fh);
	if (r != filesize)
		return -4;

	fclose(fh);

	mcio_mcEncryptContentKey(buf, ContentKey);

	printf("PS2 Memory Card Content Key\n");
	printf("target Kelf file: '%s'\n", card_kelf_path);
	printf("Kbit: ");
	for (i=0; i<16; i++)
		printf("%02x ", ContentKey[i]);
	printf("\n");
	printf("Kc  : ");
	for (i=0; i<16; i++)
		printf("%02x ", ContentKey[i+16]);
	printf("\n");

	fh = fopen(card_kelf_path, "wb");
	if (fh == NULL)
		return -5;

	r = fwrite(buf, 1, filesize, fh);
	if (r != filesize)
		return -6;

	int CK_offset = mcio_mcGetKelfContentKeyOffset(buf);
	fseek(fh, CK_offset, SEEK_SET);

	r = fwrite(ContentKey, 1, 32, fh);
	if (r != 32)
		return -6;

	fclose(fh);

	free(buf);

	return 0;
}


int main(int argc, char **argv)
{
	int r, cmd = CMD_NONE;
	char **cmd_args = NULL;

	printf(PROGRAM_NAME " v" PROGRAM_VER "\n");

	if (argc < 2) {
		print_usage(argc, argv);
		return 1;
	}
	else {
		if (!strcmp(argv[1], "--mc-info") || !strcmp(argv[1], "-i")) {
			cmd = CMD_MCINFO;
		}
		else if (!strcmp(argv[1], "--mc-free") || !strcmp(argv[1], "-f")) {
			cmd = CMD_MCFREE;
		}
		else if (!strcmp(argv[1], "--mc-image") || !strcmp(argv[1], "-img")) {
			if (argc < 3) {
				print_usage(argc, argv);
				return 1;
			}
			cmd = CMD_MCIMG;
			cmd_args = &argv[2];
		}
		else if (!strcmp(argv[1], "--mc-unformat")) {
			cmd = CMD_MCUNFORMAT;
		}
		else if (!strcmp(argv[1], "--mc-format")) {
			cmd = CMD_MCFORMAT;
		}
		else if (!strcmp(argv[1], "--list") || !strcmp(argv[1], "-ls")) {
			if (argc < 3) {
				print_usage(argc, argv);
				return 1;
			}
			cmd = CMD_LIST;
			cmd_args = &argv[2];
		}
		else if (!strcmp(argv[1], "--extract-file") || !strcmp(argv[1], "-x")) {
			if (argc < 4) {
				print_usage(argc, argv);
				return 1;
			}
			cmd = CMD_EXTRACT;
			cmd_args = &argv[2];
		}
		else if (!strcmp(argv[1], "--inject-file") || !strcmp(argv[1], "-in")) {
			if (argc < 4) {
				print_usage(argc, argv);
				return 1;
			}
			cmd = CMD_INJECT;
			cmd_args = &argv[2];
		}
		else if (!strcmp(argv[1], "--make-directory") || !strcmp(argv[1], "-mkdir")) {
			if (argc < 3) {
				print_usage(argc, argv);
				return 1;
			}
			cmd = CMD_MKDIR;
			cmd_args = &argv[2];
		}
		else if (!strcmp(argv[1], "--remove-directory") || !strcmp(argv[1], "-rmdir")) {
			if (argc < 3) {
				print_usage(argc, argv);
				return 1;
			}
			cmd = CMD_RMDIR;
			cmd_args = &argv[2];
		}
		else if (!strcmp(argv[1], "--remove") || !strcmp(argv[1], "-rm")) {
			if (argc < 3) {
				print_usage(argc, argv);
				return 1;
			}
			cmd = CMD_REMOVE;
			cmd_args = &argv[2];
		}
		else if (!strcmp(argv[1], "--file-crosslink") || !strcmp(argv[1], "-cl")) {
			if (argc < 4) {
				print_usage(argc, argv);
				return 1;
			}
			cmd = CMD_CROSSLINK;
			cmd_args = &argv[2];
		}
		else if (!strcmp(argv[1], "--content-key") || !strcmp(argv[1], "-ck")) {
			if (argc < 3) {
				print_usage(argc, argv);
				return 1;
			}
			cmd = CMD_MCCK;
			cmd_args = &argv[2];
		}
		else if (!strcmp(argv[1], "--sign-kelf") || !strcmp(argv[1], "-k")) {
			if (argc < 4) {
				print_usage(argc, argv);
				return 1;
			}
			cmd = CMD_MCKELF;
			cmd_args = &argv[2];
		}
		else {
			print_usage(argc, argv);
			return 1;
		}
	}

	r = usbd_attach_device(0x054c, 0x02ea); /* VID: 0x054c (Sony) PID: 0x02ea (PS3 MC Adaptor) */
	if (r < 0) {
		printf("Error: failed to attach USB device... (%d)\n", r);
		return 1;
	}

	r = mcio_init();
	/*if (r == sceMcResNoFormat)
		printf("Error: memory card not formated...\n");*/
	if ((r != sceMcResNoFormat) && (r < 0)) {
		printf("Error: no PS2 Memory Card detected... (%d)\n", r);
	}
	else {
		if (cmd == CMD_MCINFO) {
			r = cmd_mcinfo();
			if (r < 0)
				printf("Error: can't get MC infos... (%d)\n", r);
		}
		else if (cmd == CMD_MCFREE) {
			r = cmd_mcfree();
			if (r == sceMcResNoFormat)
				printf("Error: memory card is not formatted!\n");
			else if (r < 0)
				printf("Error: can't get MC free space... (%d)\n", r);
		}
		else if (cmd == CMD_MCIMG) {
			r = cmd_mcimg(cmd_args[0]);
			if (r < 0)
				printf("Error: can't create image file... (%d)\n", r);
		}
		else if (cmd == CMD_MCUNFORMAT) {
			r = cmd_mcunformat();
			if (r < 0)
				printf("Error: can't unformat MC... (%d)\n", r);
		}
		else if (cmd == CMD_MCFORMAT) {
			r = cmd_mcformat();
			if (r < 0)
				printf("Error: can't format MC... (%d)\n", r);
		}
		else if (cmd == CMD_LIST) {
			r = cmd_list(cmd_args[0]);
			if (r == sceMcResNoFormat)
				printf("Error: memory card is not formatted!\n");
			else if (r == sceMcResNoEntry)
				printf("Error: path '%s' not found...\n", cmd_args[0]);
			else if (r == sceMcResNotDir)
				printf("Error: path '%s' is not a directory...\n", cmd_args[0]);
			else if (r < 0)
				printf("Error: can't list directory '%s' (%d)\n", cmd_args[0], r);
		}
		else if (cmd == CMD_EXTRACT) {
			r = cmd_extract(cmd_args[0], cmd_args[1]);
			if (r == sceMcResNoFormat)
				printf("Error: memory card is not formatted!\n");
			else if (r == sceMcResNotFile)
				printf("Error: '%s' is not a file...\n", cmd_args[0]);
			else if (r < 0)
				printf("Error: can't extract file '%s'... (%d)\n", cmd_args[0], r);
		}
		else if (cmd == CMD_INJECT) {
			r = cmd_inject(cmd_args[0], cmd_args[1]);
			if (r == sceMcResNoFormat)
				printf("Error: memory card is not formatted!\n");
			else if (r < 0)
				printf("Error: can't inject file '%s'... (%d)\n", cmd_args[0], r);
		}
		else if (cmd == CMD_MKDIR) {
			r = cmd_mkdir(cmd_args[0]);
			if (r == sceMcResNoFormat)
				printf("Error: memory card is not formatted!\n");
			else if (r < 0)
				printf("Error: can't create directory '%s'... (%d)\n", cmd_args[0], r);
		}
		else if (cmd == CMD_RMDIR) {
			r = cmd_rmdir(cmd_args[0]);
			if (r == sceMcResNoFormat)
				printf("Error: memory card is not formatted!\n");
			else if (r < 0)
				printf("Error: can't remove directory '%s'... (%d)\n", cmd_args[0], r);
		}
		else if (cmd == CMD_REMOVE) {
			r = cmd_remove(cmd_args[0]);
			if (r == sceMcResNoFormat)
				printf("Error: memory card is not formatted!\n");
			else if (r < 0)
				printf("Error: can't remove file '%s'... (%d)\n", cmd_args[0], r);
		}
		else if (cmd == CMD_CROSSLINK) {
			r = cmd_crosslink(cmd_args[0], cmd_args[1]);
			if (r < 0)
				printf("Error: can't crosslink file '%s'... (%d)\n", cmd_args[0], r);
		}
		else if (cmd == CMD_MCCK) {
			r = cmd_mcck(cmd_args[0]);
			if (r < 0)
				printf("Error: can't get the Content Key... (%d)\n", r);
		}
		else if (cmd == CMD_MCKELF) {
			r = cmd_mckelf(cmd_args[0], cmd_args[1]);
			if (r < 0)
				printf("Error: can't sign the kelf... (%d)\n", r);
		}
	}

	/* release usb device */
	usbd_detach_device();

	if (r < 0)
		return 1;

	return 0;
}

